/* 
   dlanconfig.c: dlanconfig software for HomePlug devices for linux

   Copyrighted 2005 by devolo AG

   Redistribution in source form, with or without modification,
   are NOT permitted.
   This program uses explicitly confidential information
   about HomePlug standard.
*/

#include "md5.h" 
#include "dlanconfig.h"

#include <signal.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <sys/ioctl.h>
#include <net/if.h>
#include <stdio.h>
#include <sys/uio.h>
#include <fcntl.h>
#include <unistd.h>
#include <stdlib.h>
#include <netinet/in.h>
#include <errno.h>
#include <arpa/inet.h>
#include <netinet/if_ether.h> 
#include <unistd.h>
#include <string.h>
#include <sys/mman.h> 
#include <linux/if_packet.h> 
#include <ctype.h>
#include <wait.h>

/*
 *  Define
 */
#define LOCAL  '1'
#define REMOTE '2'
#define POLL '3'
#define EXIT '4'
#define GETLOCMAC  '5'

#define EXITstr   "Exit"
#define POLLstr   "List remote devices"
#define LOCALstr  "Set local network password"
#define REMOTEstr "Set remote network password"
#define PRINTOPT(a,b)  printf("%c) %s\n",a,b)
#define DOWN 1
#define UP   0

#define KEEP_MME_SETCHAR
#define KEEP_ENABLE_BACKPRESSURE

// number of retry to get localMAC
#define MAXRETRY 5

/*
 *  Global variables
 */
//static devolo_table_t* p_devTable=NULL; 
static char localifname[MAX_INAME_LEN];
static unsigned char localMAC[MACSIZE];
static char localifstatus=UP;
static devolo_table_t p_devTable; 

static char viper_setlocal_neksucceed;
static char olddevice_localnek_setting;

/*
 *   Prototypes
*/

inline int   IsSameMAC(unsigned char* mac1,unsigned char* mac2);
inline int   IsAlmostSameMAC(const void* mac1,const void* mac2); 
void  Dump_NetRespPkt(void* packet);
int   SendSetRemoteNek(char *iname,char* EK_str,unsigned char* remotePasswd);
void  GetUpdatePhyDataRates(void* packet);
inline void  PrintMAC(unsigned char* mac_addr);
void  InitNode(void);
int   WaitForNbytes(int fd,int expectedBytes,unsigned char* buf,struct timeval tv,int extraTimeSec);
int   GetNextPacket(int fileDescr,void *strPacket);
int   DoMain(char action,char* iname,char* newNek,char* remotePasswd);
void ifdown(char *);
void GetUpdateDevTable(unsigned char*  mac, unsigned char devtype);
int LocalBridgeAlreadyFound(void);
int InitIf(char *name,unsigned char* macadd);
int DoMain(char action,char* iname,char* newNek,char* remotePasswd);
int ListenForPackets(int fileDescr,int maxNumber,int maxTimeS);
void PacketReceive_callback(unsigned char *packet);
void GetLocalParExtResp(void* packet);
int SendBroadPrivateMME_DDR( char *iname);
int SendBroadPrivateMME_QIR( char *iname);
int SendSetLocalNek( char *iname,unsigned char* NEK_str);
int SendSetRemoteViperNEK( char *iname,unsigned char* NEK_str,unsigned char* DEK_str);

void ClearProgramCacheFlags(void);
void ClearRemoteDevoloTable(void);

inline double Viper_DisplayRate(double dBytesPer336usBlock);
inline double HP1_DisplayRate(double dBytes40);

/*
 * Functions
 */

/*
 *  check if the interface if down
 */
void ifdown(char *name){
  int fd;
  struct ifreq ifr;
    fd = socket(AF_INET, SOCK_DGRAM, 0);

    printf("\nBringing %s down...\n",name);
    if(fd == -1){
      perror("error opening socket");
      return;
    }
    memset(&ifr, 0, sizeof(struct ifreq));
    strncpy(ifr.ifr_name, name, IF_NAMESIZE);
    ifr.ifr_flags &= (short int)~IFF_UP;  
    if(ioctl(fd, SIOCSIFFLAGS, &ifr) <0)
      perror("Error setting flag");
}   

/*
 *  Print a formatted version of MAC address
 */
inline void PrintEncDNek(unsigned char *encdek) 
{ 
  unsigned char* mac;
  for(mac=encdek;mac<encdek+DECKSIZE-1;mac++)
    printf("%02X ",*mac);
  printf("%02X",*mac);
  fflush(stdout);
}

/*
 *  Print a formatted version of MAC address
 */
inline void PrintMAC(unsigned char *mac_addr) 
{ 
  unsigned char* mac;
  for(mac=mac_addr;mac<mac_addr+MACSIZE-1;mac++)
    printf("%02X:",*mac);
  printf("%02X",*mac);
  fflush(stdout);
}


/*
 *   Check if two MAC address are egual
 */
inline int IsSameMAC(unsigned char* mac1,unsigned char* mac2)
{
  int i;
  for(i=0;i<MACSIZE;i++)
    if(mac1[i] != mac2[i])
      return 0;
  return 1;
}


/*
 *   Send a packet to `iname' network interface
 *   return -1 on error, number of sent byte on success 
 */
int SendPacket(char *iname,unsigned char addr[],void *buffer,size_t len)
{
  int i=0;
  int retval=0;
  struct sockaddr_ll destAddr;
  struct sockaddr_ll sourceAddr;
  int sk=0;
  struct ifreq ifr;

  memset(&sourceAddr, 0, sizeof(sourceAddr));
  memset(&destAddr, 0, sizeof(destAddr));
  strcpy(ifr.ifr_name,iname);
  
  sk = socket(PF_PACKET, SOCK_DGRAM, htons(SOCK_PROTO));
  if ( sk < 0 ){
      printf("[ERROR]: SendPacket: socket unable to send packet \n");
      return -1;
  }

  if (ioctl(sk,SIOCGIFINDEX,&ifr)){
    printf("[ERROR]: SendPacket: error ioctl SIOCGIFINDEX errno=%d\n",errno);
    close(sk);
    return -1;
  }

  destAddr.sll_family = AF_PACKET;
  destAddr.sll_protocol =htons(ETHERTYPE_INTELLON); 
  destAddr.sll_ifindex= ifr.ifr_ifindex;             
  destAddr.sll_halen = 6;
  
  for(i=0;i<MACSIZE;i++)
    destAddr.sll_addr[i]=addr[i];
    
  sourceAddr.sll_family = PF_PACKET;
  sourceAddr.sll_protocol=htons(SOCK_PROTO);
  sourceAddr.sll_ifindex = ifr.ifr_ifindex;

  retval = bind(sk,(struct sockaddr *) &sourceAddr, sizeof(sourceAddr));
  if ( retval==-1){
      perror("SendPacket: bind error\n");
  }else
    retval = sendto(sk, buffer, len, 0,(struct sockaddr *) &destAddr,sizeof destAddr);
  if ( retval==-1)
    perror("SendPacket: sendto error\n");
  
#ifdef DLANCONFIG_DEBUG
  printf("Sent packet to %2x:%2x:%2x:%2x:%2x:%2x\n", 
	 addr[0], 
	 addr[1], 
	 addr[2],
	 addr[3], 
	 addr[4], 
	 addr[5]); 
#endif

  close(sk);
  return(retval);
}

/*
 *   Send a packet to `iname' network interface
 *   for ? Device Description Response
 *   return -1 on error
 */
int SendBroadPrivateMME_DDR( char *iname){  
  unsigned char mac_dest_addr[MACSIZE]; 
  private_mme_t * pHdr;
  int retval=0;

#ifdef DLANCONFIG_DEBUG
  printf("SendBroadPrivateMME_DDR\n");
#endif

  //  printf("sizeofprivate_mme_t = %d\n",sizeof(private_mme_t));

  pHdr = (private_mme_t*) malloc(MME_NET_STAT_FRAME_LEN-MME_HDR_LEN);
  
  memset(pHdr,0x00,MME_NET_STAT_FRAME_LEN-MME_HDR_LEN);
  memset(mac_dest_addr,0xff,MACSIZE);  

  pHdr->numEntries = 0x01;
  pHdr->MMEHeader = 0x02;
  pHdr->MMELength = 4; 
  pHdr->MMEOUI[0] = 0x00;
  pHdr->MMEOUI[1] = 0x04;
  pHdr->MMEOUI[2] = 0x87;
  pHdr->MID = DEVICE_DESCRIPTION_RESPONSE ;
  
  retval=SendPacket(iname,mac_dest_addr,pHdr,MME_NET_STAT_FRAME_LEN - MME_HDR_LEN);
  free(pHdr);  
  return retval;	  
}

/*
 *   Send a packet to `iname' network interface
 *   for ? QUALITY_INFO_REQUEST
 *   return -1 on error
 */
int SendBroadPrivateMME_QIR( char *iname){  
  unsigned char mac_dest_addr[MACSIZE]; 
  private_mme_t * pHdr;
  int retval=0;
  
  //  printf("sizeofprivate_mme_t = %d\n",sizeof(private_mme_t));

  pHdr = (private_mme_t*) malloc(MME_NET_STAT_FRAME_LEN-MME_HDR_LEN);
  
  memset(pHdr,0x00,MME_NET_STAT_FRAME_LEN-MME_HDR_LEN);
  memset(mac_dest_addr,0xff,MACSIZE);  

  pHdr->numEntries = 0x01;
  pHdr->MMEHeader = 0x02;
  pHdr->MMELength = 5; 
  pHdr->MMEOUI[0] = 0x00;
  pHdr->MMEOUI[1] = 0x04;
  pHdr->MMEOUI[2] = 0x87;
  pHdr->MID = QUALITY_RESPONSE;
  pHdr->data[0] = 0x01 ;// we don't know why
  retval=SendPacket(iname,mac_dest_addr,pHdr,MME_NET_STAT_FRAME_LEN - MME_HDR_LEN);
  free(pHdr);  
  return retval;	  
}


/*
 * parse the response to DDR e QIR 
 *
 */
void GetPrivateVendorSpecific(void* packet)
{
  unsigned char chMME;
  private_mme_rsp_t *p;
  p=(private_mme_rsp_t*)packet;
  chMME = p->MID & 0x7f;
  switch(chMME)
    {
    case DEVICE_DESCRIPTION_RESPONSE:      
      if(p->vendor.flags & 0x01){ //  remote viper found
	GetUpdateDevTable(p->SA, ISVIPER);
      }
      else{ // local viper found
	if(LocalBridgeAlreadyFound()){
	  if(memcmp(p_devTable.locBridge.MAC,p->SA,MACSIZE))
	    {
	      printf("\n[WARNING]: Found another Local dLAN device.\n");
	      printf("             This software cannot manage more than one local device!\n");
	      printf("             Sorry, exiting!\n");
	      exit(1);
	    }
	}
	else{
	  memcpy(p_devTable.locBridge.MAC,p->SA,MACSIZE);
	  p_devTable.locBridge.devType=ISVIPER;
	}
      }
      break;
    case QUALITY_RESPONSE:
      {
	int i,count;
	int num_entry;
	num_entry = ((unsigned int)(p->MMELength - 3))/sizeof(txqualityinfo_t);

	
	if(!IsSameMAC(p_devTable.locBridge.MAC,p->SA)){
	  break;
	}	
	
	for(i=0;i<p_devTable.devCounter;i++){   
	  for( count=0; count<num_entry;count++ ){
	    if(IsAlmostSameMAC(p_devTable.list[i].MAC,p->txData[count].Address)) {
#if OLD_ALGORITHM
	      double rate=0.0;       
	      if(p_devTable.list[i].devType == ISVIPER)
		rate = (htons(p->txData[count].BytePer40)  * 1.25);
	      else
		rate = (htons(p->txData[count].BytePer40)  * 1.0794);
	      p_devTable.list[i].rate=rate/40.0;
#else
	      if(p_devTable.list[i].devType == ISVIPER)
		p_devTable.list[i].rate=Viper_DisplayRate(htons(p->txData[count].BytePer40));
	      else
		p_devTable.list[i].rate=HP1_DisplayRate(htons(p->txData[count].BytePer40));	      
#endif
	      if(IsSameMAC(p_devTable.list[i].MAC,p->txData[count].Address))
		p_devTable.list[i].toggled=0;
	      else
		p_devTable.list[i].toggled=1;	 
	      break;
	    }
	    else{
	    }
	  }
	}
      break;
    }
    default:
      break;
    }
}

/*
 *   Send a packet to `iname' network interface
 *   for a network stats request
 */
int SendNetStatReq( char *iname ){  
  unsigned char mac_dest_addr[MACSIZE];      
  unsigned char outFrame[MME_NET_STAT_FRAME_LEN-MME_HDR_LEN];
  int retval=0;
  
#ifdef DLANCONFIG_DEBUG
  printf("SendNetStatReq\n");
#endif
  
  memset( outFrame,0x00, MME_NET_STAT_FRAME_LEN-MME_HDR_LEN); 
  memset(mac_dest_addr,0xff,MACSIZE);  
  
  outFrame[0] = 0x01;	           
  outFrame[1] = CODE_ReqNetStats;  
  outFrame[2] = 0xbb;              
  outFrame[3] = 0x80;              
  
  retval = SendPacket(iname,mac_dest_addr,outFrame,MME_NET_STAT_FRAME_LEN - MME_HDR_LEN);  
  return(retval);
}


/*
 *   Send a packet to `iname' network interface
 *   for a stats request
 */
int SendStatReq( char *iname ){
  unsigned char mac_dest_addr[MACSIZE];
  unsigned char outFrame[MME_MIN_FRAME_LEN-MME_HDR_LEN];
  int retval=0;
  
#ifdef DLANCONFIG_DEBUG
  printf("SendStatReq\n");
#endif
  
  memset(outFrame,0x00, MME_MIN_FRAME_LEN-MME_HDR_LEN);
  memset(mac_dest_addr,0xff,MACSIZE);  
	  
  outFrame[0] = 0x01;	          
  outFrame[1] = CODE_ReqStats;    
  outFrame[2] = MME_ReqStatsSize; 
  outFrame[3] = 0x0;              
  
  retval=SendPacket(iname,mac_dest_addr,outFrame,MME_MIN_FRAME_LEN-MME_HDR_LEN);  
  return(retval);
}


/*
 *   Send a packet to `iname' network interface
 *   for network statistics request
 */
int SendNetworkStatBasic( char *iname){  
  unsigned char mac_dest_addr[MACSIZE]; 
  struct mme_hdr* pHdr;
  int retval=0;
  
#ifdef DLANCONFIG_DEBUG
  printf("SendNetworkStatBasic\n");
#endif
  
  pHdr = (struct mme_hdr*) malloc(MME_NET_STAT_FRAME_LEN-MME_HDR_LEN);
  
  memset(mac_dest_addr,0xff,MACSIZE);
  memset(pHdr,0x00,MME_NET_STAT_FRAME_LEN-MME_HDR_LEN);
  
  pHdr->MCTRL =   0x01 ;         
  pHdr->MEHDR =   CODE_ReqNetStats;
  pHdr->MELEN =   0xbb;            
  pHdr->NETW_CTRL = 0x80 & 0x7F;   

  retval=SendPacket(iname,mac_dest_addr,pHdr,MME_NET_STAT_FRAME_LEN - MME_HDR_LEN);
  free(pHdr);  
  return retval;	  
}


/*
 *   Send a packet to `iname' network interface
 *   for local param request
 */
int SendLocalParExt( char *iname) {  
  unsigned char mac_dest_addr[MACSIZE];  
  struct mme_hdr* pHdr;
  int retval=0;

#ifdef DLANCONFIG_DEBUG
  printf("SendLocalParExt\n");
#endif

  pHdr = (struct mme_hdr*) malloc(MME_NET_STAT_FRAME_LEN-MME_HDR_LEN);
  memset(mac_dest_addr,0xff,MACSIZE); 
  memset(pHdr,0x00,MME_NET_STAT_FRAME_LEN-MME_HDR_LEN);
  
  pHdr->MCTRL = 0x1 ;                 
  pHdr->MEHDR = CODE_SetLocPar ;      
  pHdr->MELEN = MME_LocalParExtSize;  
  
  retval=SendPacket(iname,mac_dest_addr,pHdr,MME_NET_STAT_FRAME_LEN - MME_HDR_LEN);
  free(pHdr);
  return retval;	  
}


/*
 *  Function that update physical 
 *  data rates
 */
void GetUpdatePhyDataRates(void* packet)
{
  struct API_ReqNetStatsResp* p;
  int    i,count;
  double rate=0.0;
  
  p=(struct API_ReqNetStatsResp*)packet;
  
  for(i=0;i<p_devTable.devCounter;i++){   
    for( count=0; count<NWSIZE;count++ ){ 
      
      if(IsAlmostSameMAC(p_devTable.list[i].MAC,p->NW[count].DA)) {  
	rate = ((p->NW[count].BYTES40) / 42.0);
	rate = ( M_RATE * rate ) + B_RATE;
	p_devTable.list[i].rate=rate;
	
	if(IsSameMAC(p_devTable.list[i].MAC,p->NW[count].DA))
	  p_devTable.list[i].toggled=0;
	else
	  p_devTable.list[i].toggled=1;	 
	break;
      }
    }
  }
}

/*
 *  Checks if two MAC are equal apart from last bit 
 */
inline int IsAlmostSameMAC(const void* mac1,const void* mac2)
{
  char* macA;
  char* macB;
  int flag=0;

  macA=(char*)mac1;
  macB=(char*)mac2;
  
  flag=((macA[0]|0x01)==(macB[0]|0x01));  
  return( flag && macA[1]==macB[1] && macA[2]==macB[2] 
	  && macA[3]==macB[3] && macA[4]==macB[4] && macA[5]==macB[5]);
}


void ClearRemoteDevoloTable(void)
{
  int i;

  for(i=0;i<p_devTable.devCounter;i++){
    memset(p_devTable.list[i].MAC,0,MACSIZE); 
    memset(p_devTable.list[i].DEK,0,DECKSIZE); 
    memset(p_devTable.list[i].NEK,0,DECKSIZE);
    p_devTable.list[i].rate=0.0;                
    p_devTable.list[i].toggled=0;              
    p_devTable.list[i].devType=0;
    p_devTable.list[i].isHere=0;
  }
  p_devTable.devCounter=0;  
}

void ClearProgramCacheFlags(void)
{
  int i;
  for(i=0;i<p_devTable.devCounter;i++){
    p_devTable.list[i].isHere=0;
  }
}

/*
 *  Function that print the list of 
 *  connected devices
 */
void PrintNode(char* iname)
{
  unsigned int i,j=1;
  unsigned int slow_counter = 0;
  unsigned int viper_counter = 0;
  double aggRate_slow=0.0;  
  double aggRate_viper=0.0;  
  
  printf("\n");
  if(p_devTable.devCounter<=0){
    printf("No remote devices found\n");
    return;
  }
 
  for(i=0;i<p_devTable.devCounter;i++)
    {
      // aggiungeere la limitazione ai toggled
      if(p_devTable.list[i].devType==ISVIPER)
	{
	  viper_counter++;
	  aggRate_viper += p_devTable.list[i].rate;
	}
      else
	{
	  slow_counter++;
	  aggRate_slow += p_devTable.list[i].rate;
	}
    }
  
  if(slow_counter)
    aggRate_slow /= slow_counter;
  
  if(viper_counter)
    aggRate_viper /= viper_counter;

  printf("Remote devices:\n");

  /* bug 1249 */
  /*  printf("\tNum.\tMAC address       \tMbps\tCache\tisViper\tProgramCache\n"); */
  printf("\tNum.\tMAC address       \tMbps\tCache\tHighspeed\n");
/*   printf("\tNum.\tMAC address       \tMbps\tHighspeed\n");  */
  for(i=0;i<p_devTable.devCounter;i++){
    //    if(!p_devTable.list[i].toggled) {
      printf("\t%d\t",j++);
      PrintMAC(p_devTable.list[i].MAC);
      if (p_devTable.list[i].rate > 0.0)
	printf("\t%1.3f",p_devTable.list[i].rate);
      else /* Bug 1251 */
	printf("\t N/A ");
      if(p_devTable.list[i].toggled)
	printf("\t  *");
      else
	printf("\t   ");
      if(p_devTable.list[i].devType==ISVIPER)
	printf("\t  #");
      else
	printf("\t   ");    
      /* bug 1249 */
      /*       if(p_devTable.list[i].isHere) */
      /* 	printf("\t  found"); */
      /*       else */
      /* 	printf("\t  no found"); */
      printf("\n");
      //    }
  }
  printf("\n");
  /* bug 1249 */
  if(slow_counter)
    printf("Physical average data rate (Normal devices   )= %1.3f Mbps\n",aggRate_slow);

  if(viper_counter)
    printf("Physical average data rate (Highspeed devices)= %1.3f Mbps\n",aggRate_viper);

  fflush(stdout);
}


/*
 *   Usage
 */
void usage(char *progname){
  printf("Usage: %s interface\n",progname);
  printf("Example: %s dlanusb0\n\n",progname);
}

/*
 *  Send packet to set remote nek
 */
int SendSetRemoteNek(char *iname,char* EK_str,unsigned char* remotePasswd){ 
  unsigned char mac_dest_addr[MACSIZE]; 
  struct mme_double* mme_d; 
  struct mme_triple* mme_tr;
  unsigned char* pKey;
  int retval=0;
 
#ifdef DLANCONFIG_DEBUG
  printf("SendSetRemoteNek\n");
#endif

  mme_d=(struct mme_double*) malloc(MME_MIN_FRAME_LEN);
  mme_tr=(struct mme_triple*) malloc(MME_MIN_FRAME_LEN);
  
  memset(mac_dest_addr,0xff,MACSIZE);
  memset(mme_d,0x00,MME_MIN_FRAME_LEN) ; 
  
  mme_d->CTRL  = 0x2 ;   
  mme_d->HDR1  = 0x04;   
  mme_d->LEN1  = 0x09;   
  mme_d->EKS   = 0x00;   

  pKey = MD5HomePlugKey(remotePasswd);
  memcpy(mme_d->NEK,pKey,DECKSIZE);  

#ifdef KEEP_MME_SETCHAR
  /* change characteristics.:  removed to fix bug 1082....*/
  mme_d->HDR2      = 0x1F;
  mme_d->LEN2      = 0x03;
  mme_d->TXCHAR[0] = 0x58;   
  mme_d->TXCHAR[1] = 0x84;
#ifdef KEEP_ENABLE_BACKPRESSURE
  mme_d->TXCHAR[1] |= 0x02;
#endif   
  mme_d->TXCHAR[2] = 0x00;
  retval=SendPacket(iname,mac_dest_addr,mme_d,MME_MIN_FRAME_LEN-MME_HDR_LEN); 
#else
  retval=SendPacket(iname,mac_dest_addr,mme_d,MME_MIN_FRAME_LEN-MME_HDR_LEN-5); 
#endif


  if(retval<=0){
#ifdef DLANCONFIG_DEBUG
    printf("[ERROR]: SendSetRemoteNek: sending first packet\n");
#else
    printf("Please check that the dLAN adapter's security ID is correct or/and that it is NOT in standby mode.\n");
#endif
    free(mme_d);
    free(mme_tr);
    return(retval);
  }
  
  memset(mac_dest_addr,0xff,MACSIZE); 
  memset(mme_d,0x00,MME_MIN_FRAME_LEN - MME_HDR_LEN);

  // added check response  mme_d->CTRL  = 0x01;  
  mme_d->CTRL  = 0x02;  
  mme_d->HDR1  = 0x04;  
  mme_d->LEN1  = 0x09;  
  mme_d->EKS   = 0x01;  

  pKey = MD5HomePlugKey(EK_str); 
  memcpy(mme_d->NEK,pKey,DECKSIZE);  
  
  // added check response
  mme_d->HDR2   =   CODE_ReqStats;
  mme_d->LEN2   =   MME_ReqStatsSize; 
  mme_d->TXCHAR[0] = 0x00;   

  retval=SendPacket(iname,mac_dest_addr,mme_d,MME_MIN_FRAME_LEN-MME_HDR_LEN); 
  if(retval<=0){
#ifdef DLANCONFIG_DEBUG
    printf("[ERROR]: SendSetRemoteNek: sending second packet\n");
#else
    printf("Please check that the dLAN adapter's security ID is correct or/and that it is NOT in standby mode.\n");
#endif
    free(mme_d);
    free(mme_tr);
    return(retval);
  }
  
  memset(mac_dest_addr,0xff,MACSIZE); 
  memset(mme_tr,0x00,MME_MIN_FRAME_LEN) ; 

  mme_tr->CTRL  = 0x03;   
  mme_tr->HDR1  = 0x04;   
  mme_tr->LEN1  = 0x09;   
  mme_tr->EKS1  = 0x00;   
  memcpy(mme_tr->NEK1,p_devTable.locBridge.DEK,DECKSIZE);

  mme_tr->HDR2  = 0x04; 
  mme_tr->LEN2  = 0x09; 
  mme_tr->EKS2  = 0x01;
  memcpy(mme_tr->NEK2,p_devTable.locBridge.NEK,DECKSIZE);


#ifdef KEEP_MME_SETCHAR
  /* change characteristics.:  removed to fix bug 1082....*/
  mme_tr->HDR3  = 0x1F;
  mme_tr->LEN3  = 0x03;
  mme_tr->TXCHAR[0] = 0x58;  
  mme_tr->TXCHAR[1] = 0x84;
#ifdef KEEP_ENABLE_BACKPRESSURE
  mme_d->TXCHAR[1] |= 0x02;
#endif  
  mme_tr->TXCHAR[2] = 0x01;
  retval=SendPacket(iname,mac_dest_addr,mme_tr,MME_MIN_FRAME_LEN-MME_HDR_LEN); 
#else
  retval=SendPacket(iname,mac_dest_addr,mme_tr,MME_MIN_FRAME_LEN-MME_HDR_LEN-5); 
#endif


  
  if(retval<=0){
#ifdef DLANCONFIG_DEBUG
    printf("[ERROR]: SendSetRemoteNek: sending third packet\n");
#else
    printf("Please check that the dLAN adapter's security ID is correct or/and that it is NOT in standby mode.\n");
#endif
    free(mme_d);
    free(mme_tr);
    return(retval);
  }

  memset(mac_dest_addr,0xff,MACSIZE); 
  memset(mme_d,0x00,MME_MIN_FRAME_LEN-MME_HDR_LEN);
  mme_d->CTRL  = 0x2 ;  
  mme_d->HDR1  = 0x04;  
  mme_d->LEN1  = 0x09;  
  mme_d->EKS   = 0x01; 
  memcpy(mme_d->NEK,p_devTable.locBridge.NEK,DECKSIZE);
  
#ifdef KEEP_MME_SETCHAR
  /* change characteristics.:  removed to fix bug 1082....*/
  mme_d->HDR2      = 0x1F;
  mme_d->LEN2      = 0x03;
  mme_d->TXCHAR[0] = 0x58;   
  mme_d->TXCHAR[1] = 0x84;
#ifdef KEEP_ENABLE_BACKPRESSURE
  mme_d->TXCHAR[1] |= 0x02;
#endif
  mme_d->TXCHAR[2] = 0x01;
   
  retval=SendPacket(iname,mac_dest_addr,mme_d,MME_MIN_FRAME_LEN-MME_HDR_LEN); 
#else
  retval=SendPacket(iname,mac_dest_addr,mme_d,MME_MIN_FRAME_LEN-MME_HDR_LEN-5); 
#endif

  if(retval<=0){
#ifdef DLANCONFIG_DEBUG
   printf("[ERROR]: SendSetRemoteNek: sending fourth packet\n");
#else
    printf("Please check that the dLAN adapter's security ID is correct or/and that it is NOT in standby mode.\n");
#endif
   free(mme_d);
   free(mme_tr);
   return(retval);
  }
  free(mme_d);
  free(mme_tr);
  return retval;
}


/*
 *  Send packet to poll network
 */
int API_NetwPoll(char* iname)
{
  int count;
  int retval=0;
  
  for(count=0;count<POLL_RETRY;count++){
    retval+=SendStatReq(iname);
    sleep(INTER_FRAME_DELAY);
    retval+=SendNetStatReq(iname);
  }
  return retval;
}


/*
 *  Print main menu
 */
inline void printMenu(){
  printf(  "\nMain menu:\n");
  PRINTOPT(LOCAL,LOCALstr);
  PRINTOPT(REMOTE,REMOTEstr);
  PRINTOPT(POLL,POLLstr);
  PRINTOPT(EXIT,EXITstr);
  printf("\nChoose a command number: ");
}  
/*
* close child and associated filedescriptors
*/
inline void CloseAll(int fd,pid_t pid){
  //  free(p_devTable);
  close(fd);
  kill(pid,15);
  wait(NULL);
}

/*
 *   Wait until expectedBytes arrive
 */
int   WaitForNbytes(int fd,int expectedBytes,unsigned char* buf,struct timeval input_tv,int extraTimeSec){
  fd_set rfds;
  int arrivedBytes=0;
  int ready=0;
  struct timeval MyTimeval,tv; 

  FD_ZERO(&rfds);
  FD_SET(fd, &rfds);
  arrivedBytes=0;
  tv.tv_sec=input_tv.tv_sec;  
  tv.tv_usec=input_tv.tv_usec;
  
  ready=select(fd+1, &rfds, NULL, NULL, &tv);
  if(ready)   
    arrivedBytes=read(fd,buf,expectedBytes);
  else        
    return 0;
  
  if(arrivedBytes==expectedBytes)
    {
      //      printf("[INFO]: WaitForNbytes: read n=%d bytes\n",arrivedBytes);
      
      return 1;
    } 
  if(arrivedBytes==0){ 
    printf("[ERROR]: WaitForNbytes: no bytes to read\n");
    return -1;
  }
  
  if(arrivedBytes<expectedBytes) { 
    FD_ZERO(&rfds);
    FD_SET(fd, &rfds);
    MyTimeval.tv_sec=3;
    MyTimeval.tv_usec=0;  
    ready=select(fd+1, &rfds, NULL, NULL, &MyTimeval);
    if(ready)
      arrivedBytes+=read(fd,&(buf[arrivedBytes]),expectedBytes-arrivedBytes);
    if(arrivedBytes==expectedBytes) { 
      return(1); 
    }else{
      printf("[ERROR]: WaitForNbytes: no bytes to read\n");
      return(-1);
    }
  }
  printf("[ERROR]: WaitForNbytes: incomplete reading\n");
  return (-1); 
}


/*
 *   Get succesive packet
 */
int GetNextPacket(int fileDescr,void *strPacket)
{
  struct timeval tv;
  int retval=0;
  unsigned char strHeader[PACKET_HEADERSIZE]="";
  int packLen=0;
  
  tv.tv_sec=SELECT_TIMEOUT_S;
  tv.tv_usec=SELECT_TIMEOUT_US;
  
  retval=WaitForNbytes(fileDescr,PACKET_HEADERSIZE-1,strHeader,tv,EXTRA_WAIT_TIME_S);

  switch(retval){
  case -1:
    printf("[ERROR]: GetNextPacket: broken protocol, incomplete Header\n");
    return(-1);
  case 0:
    return 0;
  case 1:
    strHeader[PACKET_HEADERSIZE-1]='\0';
    packLen=atoi(strHeader);
    memset(strPacket,0,MAX_PACK_LEN);
    retval=WaitForNbytes(fileDescr,packLen,strPacket,tv,EXTRA_WAIT_TIME_S);
    
    if(retval==-1){
      printf("[ERROR]: GetNextPacket: broken protocol, incomplete packet!\n");
      return (-1);
    }
    
    if(retval==0) {
      printf("[ERROR]: GetNextPacket: broken protocol, header present but no packet available!\n");
      return(0);
    }   
    {
      /**********************/
      int c=0;
      struct API_ReqNetStatsResp* p;
      p=(struct API_ReqNetStatsResp*)strPacket;
      c++;
      /*******************/
    }
    return(1);
  }
  
  return -1;
}

/*
 *  Main
 */
int main(int argc, char *argv[]){
  //  char iname[MAX_INAME_LEN]="";
  char buff[256];
  int cont=TRUE;

  printf("\n<-- dlanconfig software v%s -->\n\n",DLANCONFIG_VERSION);
 
  /*check args*/
  if(argc!=2) {
    usage(argv[0]); 
    return 0;
  }
  if(!strncmp(argv[1],"--h",3) || !strncmp(argv[1],"-h",2)) {
    usage(argv[0]); 
    return 0;
  }
  strncpy(localifname,argv[1],MAX_INAME_LEN);
  
  /* check root proviledges */
  if(getuid()!=0){
    printf("\nTo run dlanconfig software you must have root priviledges.\n"); 
    fflush(stdout);  
    return 0; 
  }

  /* check interface*/
  if(InitIf(localifname,localMAC)){
    printf("\nSorry, exiting!\n");
    return 1;
  }  
  

  /* looking for local dLAN device*/
  if(DoMain(GETLOCMAC,localifname,NULL,NULL)){
    printf("\nSorry, exiting!\n");
    return 1;
  }
  
  printf("\nLocal dLAN device has MAC address: ");
  PrintMAC(p_devTable.locBridge.MAC);
#ifdef DLANCONFIG_DEBUG
  /* bug 1250 */
    printf("\n\t\t\t enc. NEK: ");
    PrintEncDNek(p_devTable.locBridge.NEK);
    printf("\n\t\t\t enc. DEK: ");
    PrintEncDNek(p_devTable.locBridge.DEK);
#endif
  printf("\n");

  while (cont){
    printMenu();
    fgets((char *)buff,256,stdin);
    switch((char)buff[0]){
    case EXIT:
      if(localifstatus == DOWN)
	ifdown(localifname);
      cont=FALSE;
      break;
    case POLL:
      {
/*  	ClearProgramCacheFlags();  */
/*      bug 1111 */
	ClearRemoteDevoloTable();
	if(!DoMain(POLL,localifname,NULL,NULL)){
	  printf("\nLocal dLAN device has MAC address ");
	  PrintMAC(p_devTable.locBridge.MAC);
	  printf("\n");
	  PrintNode(localifname);
	}
      }
      break;
    case LOCAL:
      {
	char netpasswd[PASSWDMAXLEN]="";
	printf("\n%s\n",LOCALstr);
	printf("New network password to set: ");
	fgets(netpasswd, PASSWDMAXLEN, stdin);
	netpasswd[strlen(netpasswd)-1]='\0';
	if(!DoMain(LOCAL,localifname,netpasswd,NULL))
	  printf("OK\n");
      }
      break;
      case REMOTE: 
      {
	int tryfor=4;
	char netpasswd[PASSWDMAXLEN]="";
	char devpasswd[PASSWDMAXLEN]="";
	printf("\n%s\n",REMOTEstr);
	printf("New network password to set: ");
	fgets(netpasswd, PASSWDMAXLEN, stdin);
	netpasswd[strlen(netpasswd)-1]='\0';

	while(tryfor){
	  /* bug 1259 */
	  /*	  printf("Remote device password: "); */
	  printf("Security ID: "); 
	  fgets(devpasswd, PASSWDMAXLEN, stdin);
	  devpasswd[strlen(devpasswd)-1]='\0'; // delete new line
	  if(strlen(devpasswd)==19 &&
	     devpasswd[4] == '-' &&
	     devpasswd[9] == '-' &&
	     devpasswd[14] == '-')
	    break;
	  tryfor--;
	  /* bug 1259 */
	  /* printf("error: Remote device password has the form: XXXX-XXXX-XXXX-XXXX\n"); */
	  printf("error: Security ID has the form: XXXX-XXXX-XXXX-XXXX\n"); 
	}
	if(tryfor){/*bug 1275 */
	  printf("\nSetting local network password. ");
	  if(!DoMain(LOCAL,localifname,netpasswd,NULL)){
	      printf("OK\n");
	      printf("Setting remote network password. ");
	      if(!DoMain(REMOTE,localifname,netpasswd,devpasswd)){
		printf("OK\n");
		printf("Rescanning the network. ");
		// fix no bug reported
		ClearRemoteDevoloTable();
		if(!DoMain(POLL,localifname,NULL,NULL)){
		  printf("\nLocal dLAN device has MAC address ");
		  PrintMAC(p_devTable.locBridge.MAC);
		  printf("\n");
		  PrintNode(localifname);
		}
	      }
	      printf("\nWARNING: your local password was set to %s\n",netpasswd);
	  }
	}
	
      }
      break;
    }
  }
  return 0;
}

/*
 *   Function that forks the processes
 */
int DoMain(char action,char* iname,char* newNek,char* remotePasswd){
  pid_t childpid;
  int retval=0;
  int mypipe[2];  
  int i=0;  
  //  unsigned char nullMAC[MACSIZE]={0};
  int retry;
  char* myArgv[3];

  myArgv[0]=SNIFFING_CHILD_PROCESS;
  myArgv[1]=iname;
  myArgv[2]=NULL;


  /*
   *   Create pipe
   */
  errno=0;
  if(pipe(mypipe)!=0){    
    printf("[ERROR]: DoMain: Error creating the pipe\n");
    strerror(errno);
    return -1;
  }


  /* 
   *  fork
   */
  if((childpid=fork())==0){ 

    close(mypipe[0]);
    dup2(mypipe[1],STDOUT_FILENO); 
    close(mypipe[1]);
    errno=0;
    execvp(SON_PROCESS_NAME,(char**)myArgv);
    fprintf(stderr,"cf: %s return\n",SNIFFING_CHILD_PROCESS);
    strerror(errno);
    fflush(stdout);
  }
  
  close(mypipe[1]);
  
  printf("\nPlease wait.");fflush(stdout);    
  retval=1;
  switch(action){
  case GETLOCMAC:
    //looking for viper
    for(retry=MAXRETRY; retry ; retry--){
      // send private request looking for viper devices
      retval=SendBroadPrivateMME_DDR(iname);
      if(retval<0){
	printf("\nUnable to send packets!\n");
	CloseAll(mypipe[0],childpid);
	return 1;
      }
      usleep(INTER_FRAME_DELAY_USEC); // ORIGINAL 1 SEC TRY TO REDUCE IT
      printf(".");fflush(stdout);
      // waiting some reply 
      retval = ListenForPackets(mypipe[0],MAX_PACKET_TO_PARSE,WAIT_AFTER_NOACTIVITY);
      if(retval<0){
	printf("\nUnable to receive packets!\n");
	CloseAll(mypipe[0],childpid);
	return 1;
      }
      printf(".");fflush(stdout);
      // now looking for dlan 14Mb devices
      // I retry more times in case of multiples local bridges!
      for(i=0;i<LOCAL_POLL_RETRY;i++){
	retval=SendLocalParExt(iname);
	if(retval<0){
	  printf("\nUnable to send packets!\n");
	  CloseAll(mypipe[0],childpid);
	  return 1;
	}
	usleep(INTER_FRAME_DELAY_USEC);
	printf(".");fflush(stdout);
	retval = ListenForPackets(mypipe[0],MAX_PACKET_TO_PARSE,WAIT_AFTER_NOACTIVITY);
	if(retval<0){
	  printf("\nUnable to receive packets!\n");
	  CloseAll(mypipe[0],childpid);
	  return 1;
	}
	printf(".");fflush(stdout);
      }
      if(LocalBridgeAlreadyFound()){
	CloseAll(mypipe[0],childpid);
	printf("\nLocal bridge found\n");
	return 0;
      }      
      // Now I check if I have found one local bridge
      // ??
    }
    CloseAll(mypipe[0],childpid);
    printf("No local device found");
    return 1; 
    break;
  case POLL:
    // looking for device on dLAN...
    retval=SendStatReq(iname);
    if(retval<0){
      printf("\nUnable to send packets!\n");
      CloseAll(mypipe[0],childpid);
      return 1;
    }
    sleep(INTER_FRAME_DELAY);
    printf(".");fflush(stdout);
    retval = ListenForPackets(mypipe[0],MAX_PACKET_TO_PARSE,WAIT_AFTER_NOACTIVITY);
    if(retval<0){
      printf("\nUnable to receive packets!\n");
      CloseAll(mypipe[0],childpid);
      return 1;
    }
    printf(".");fflush(stdout);
    usleep(INTER_FRAME_DELAY_USEC); // ORIGINAL 1 SEC TRY TO REDUCE IT
    // looking for device viper on dlan just to add viper flags!
    retval=SendBroadPrivateMME_DDR(iname);
    if(retval<0){
      printf("\nUnable to send packets!\n");
      CloseAll(mypipe[0],childpid);
      return 1;
    }
    usleep(INTER_FRAME_DELAY_USEC); // ORIGINAL 1 SEC TRY TO REDUCE IT
    printf(".");fflush(stdout);
    // waiting some reply 
    retval = ListenForPackets(mypipe[0],MAX_PACKET_TO_PARSE,WAIT_AFTER_NOACTIVITY);
    if(retval<0){
      printf("\nUnable to receive packets!\n");
      CloseAll(mypipe[0],childpid);
      return 1;
    }
    if(p_devTable.locBridge.devType == ISOLDDEVICE){ // get rate with old device as local bridge.
      retval=SendNetStatReq(iname);
      if(retval<0){
	printf("\nUnable to send packets!\n");
	CloseAll(mypipe[0],childpid);
	return 1;
      }
    }
    else if(p_devTable.locBridge.devType == ISVIPER)
      {
	retval =SendBroadPrivateMME_QIR(iname);
	if(retval<0){
	  printf("\nUnable to send packets!\n");
	  CloseAll(mypipe[0],childpid);
	  return 1;
	}
      }
    else
      {
	printf("\nlocal Bridge type is not recognized!\n");
	CloseAll(mypipe[0],childpid);
	return 1;	
      }
    printf(".");fflush(stdout);
    sleep(INTER_FRAME_DELAY);
    printf(".");fflush(stdout);
    retval = ListenForPackets(mypipe[0],MAX_PACKET_TO_PARSE,WAIT_AFTER_NOACTIVITY);
    if(retval<0){
      printf("\nUnable to receive packets!\n");
      CloseAll(mypipe[0],childpid);
      return 1;
    }
    printf(".");fflush(stdout);
    //    return 0;
    break;
  case LOCAL:
    { 
      viper_setlocal_neksucceed=0;
      olddevice_localnek_setting=1;
      retval = SendSetLocalNek( localifname,newNek);
      if(retval<0){
	printf("\nUnable to send packets!\n");
	CloseAll(mypipe[0],childpid);
	return 1;
      }
      printf(".");fflush(stdout);
      usleep(INTER_FRAME_DELAY_USEC);
      printf(".");fflush(stdout); 
      if(p_devTable.locBridge.devType==ISVIPER){
	retval = ListenForPackets(mypipe[0],MAX_PACKET_TO_PARSE,WAIT_AFTER_NOACTIVITY);
	printf(".");fflush(stdout);
	if(retval<0){
	  retval = SendSetLocalNek(localifname,newNek);
	  if(retval<0){
	    printf("\nUnable to send packets!\n");
	    CloseAll(mypipe[0],childpid);
	    return 1;
	  }
	  usleep(INTER_FRAME_DELAY_USEC);
	  printf(".");fflush(stdout); 
	  retval = ListenForPackets(mypipe[0],MAX_PACKET_TO_PARSE,WAIT_AFTER_NOACTIVITY);
	  if(retval<0){
	    printf("\nUnable to receive packets!\n");
	    CloseAll(mypipe[0],childpid);
	    return 1;
	  }
	}
	if(viper_setlocal_neksucceed==0)
	  {
	    printf("\nUnable to change local password!\n");
	    CloseAll(mypipe[0],childpid);
	    return 1;
	  }
      }
      else{
	// read localparam and check new enc. NEK
	retval=SendLocalParExt(iname);
	if(retval<0){
	  printf("\nUnable to send packets!\n");
	  CloseAll(mypipe[0],childpid);
	  return 1;
	}
	usleep(INTER_FRAME_DELAY_USEC);
	printf(".");fflush(stdout);
	usleep(INTER_FRAME_DELAY_USEC);
	printf(".");fflush(stdout);
	retval = ListenForPackets(mypipe[0],MAX_PACKET_TO_PARSE,WAIT_AFTER_NOACTIVITY);
	if(olddevice_localnek_setting==0){
	  if(memcmp(p_devTable.locBridge.NEK,MD5HomePlugKey(newNek),DECKSIZE)){
	    printf("\nUnable to change local password!\n");
	    CloseAll(mypipe[0],childpid);
	    return 1;
	  }
	}
	printf(".");fflush(stdout);
	retval = 1; //force correct return
      }
    }
    //    return 0;    
    break;
  case REMOTE:
    if(p_devTable.locBridge.devType==ISVIPER){
      viper_setlocal_neksucceed=0;
      retval = SendSetRemoteViperNEK(iname,newNek,remotePasswd);
      if(retval<0){
	printf("\nUnable to send packets!\n");
	CloseAll(mypipe[0],childpid);
	return 1;
      }
      printf(".");fflush(stdout);
      usleep(INTER_FRAME_DELAY_USEC);
      printf(".");fflush(stdout); 
      retval = ListenForPackets(mypipe[0],MAX_PACKET_TO_PARSE,WAIT_AFTER_NOACTIVITY);
      printf(".");fflush(stdout);
      if(retval<0){
	retval = SendSetRemoteViperNEK(iname,newNek,remotePasswd);
	if(retval<0){
	  printf("\nUnable to send packets!\n");
	  CloseAll(mypipe[0],childpid);
	  return 1;
	}
	usleep(INTER_FRAME_DELAY_USEC);
	printf(".");fflush(stdout); 
	retval = ListenForPackets(mypipe[0],MAX_PACKET_TO_PARSE,WAIT_AFTER_NOACTIVITY);
	if(retval<0){
	  printf("\nUnable to receive packets!\n");
	  CloseAll(mypipe[0],childpid);
	  return 1;
	}
      }
      if(viper_setlocal_neksucceed==0)
	{
	  printf("\nUnable to change remote password!\n");
	  printf("Please check that the dLAN adapter's security ID is correct or/and that it is NOT in standby mode.\n");
	  CloseAll(mypipe[0],childpid);
	  return 1;
	}      
    }
    else{//olddevice
      viper_setlocal_neksucceed=0;
      retval=SendSetRemoteNek(iname,newNek,remotePasswd);
      if(retval<0){
	printf("\nUnable to send packets!\n");
	CloseAll(mypipe[0],childpid);
	return 1;
      }
      printf(".");fflush(stdout); 
      usleep(INTER_FRAME_DELAY_USEC);
      printf(".");fflush(stdout); 
      retval = ListenForPackets(mypipe[0],MAX_PACKET_TO_PARSE,WAIT_AFTER_NOACTIVITY);
      if(retval<0){	
	retval=SendSetRemoteNek(iname,newNek,remotePasswd);
	if(retval<0){
	  printf("\nUnable to send packets!\n");
	  CloseAll(mypipe[0],childpid);
	  return 1;
	}
	printf(".");fflush(stdout); 
	usleep(INTER_FRAME_DELAY_USEC);
	printf(".");fflush(stdout); 
	retval = ListenForPackets(mypipe[0],MAX_PACKET_TO_PARSE,WAIT_AFTER_NOACTIVITY);
	if(retval<0){	
	  printf("\nUnable to receive packets!\n");
	  CloseAll(mypipe[0],childpid);
	  return 1;
	}
      }
      usleep(INTER_FRAME_DELAY_USEC);
      if(viper_setlocal_neksucceed==0)
	{
	  printf("\nUnable to change remote password!\n");
	  printf("Please check that the dLAN adapter's security ID is correct or/and that it is NOT in standby mode.\n");
	  CloseAll(mypipe[0],childpid);
	  return 1;
	}
    }
    break;
  }

  if(retval>=0){ 
    CloseAll(mypipe[0],childpid);
    return 0;
  }else{
    printf("Operation FAILED\n");  
    CloseAll(mypipe[0],childpid);
    return 1;
  }
}


/*
  checks for the iff_up flag and then bring interface up
  if necessary. brings interface MAC
  param in  name: interface name
  param out macadd: MAC address
  return: 1 on error, 0 on successfull
 */
int InitIf(char *name,unsigned char* macadd){

  int fd;
  struct ifreq ifr;
  
  if(!macadd){
    printf("Oops. Internal error %s.\n",__FUNCTION__);
    return 1;
  }

  fd = socket(AF_INET, SOCK_DGRAM, 0);
  
  if(fd == -1){
    perror("\n Error opening socket: ");
    return 1;
  }
  
  memset(&ifr, 0, sizeof(struct ifreq));
  strncpy(ifr.ifr_name, name, IF_NAMESIZE);
  
  if(ioctl(fd, SIOCGIFFLAGS, &ifr)<0){
    /* necessary to let kernel modprobe driver
       and eventually ifup the interface */
    sleep(1);
    if(ioctl(fd, SIOCGIFFLAGS, &ifr)<0){
      perror(NULL);
      close(fd);
      return 1;
    }
  }
  
  
  if ((ifr.ifr_flags & IFF_UP) == 0){
    localifstatus=DOWN;
    printf("Interface %s is down. Bringing %s up..",name,name);
    fflush(stdout);
    ifr.ifr_flags |= IFF_UP;
    //ifup interface
    if(ioctl(fd, SIOCSIFFLAGS, &ifr) <0){
      perror("\nError bringing up: ");
      close(fd);
      return 1;
    }
    printf(".OK\n");
  }
  
  //get mac address
  if ( ioctl(fd,SIOCGIFHWADDR, &ifr) == -1 ){
    perror("\nError getting MAC address: ");
    close(fd);
    return 1;
  }
  memcpy(macadd,ifr.ifr_hwaddr.sa_data,MACSIZE);
  close(fd);
  return 0;
}


/*
 *   Listen until new packets arrive
 *   return number of received packet, -1 on error
 */
int ListenForPackets(int fileDescr,int maxNumber,int maxTimeS)
{
  int receivedCount=0;
  int timeoutCount=0;
  int timeoutN=0;
  int retval=0;
  unsigned char buffer[MAX_PACK_LEN];  


  timeoutN=(1000000*maxTimeS)/SELECT_TIMEOUT_US;
  while(receivedCount<maxNumber && timeoutCount<timeoutN) {
    retval=GetNextPacket(fileDescr,buffer);
    switch(retval){
    case -1:
      printf("[ERROR]: ListenForPackets: get next packet = %d\n",retval );   
      return -1;
    case 0:
      timeoutCount++;  
      break;
    case 1:
      PacketReceive_callback(buffer); 
      receivedCount++;    
      break;
    }
  }
  return(receivedCount);

}

/*
 *   Parser of received packet 
 */
void PacketReceive_callback(unsigned char *packet)
{
  struct MME_PackedHdr* pHdr;

  if(packet == NULL)
    return;
  pHdr = (struct MME_PackedHdr*)packet;

#ifdef DLANCONFIG_DEBUG
  printf("Received packet with pHdr->MEHDR:%d\n",pHdr->MEHDR); 
#endif

  if(!IsSameMAC(pHdr->DA,localMAC) || pHdr->Ethertype!=htons(ETHERTYPE_INTELLON)) {
#ifdef DLANCONFIG_DEBUG
    printf("Packet is not for us or is not ETHERTYPE_INTELLON\n"); 
#endif
    return;
  }

  if((pHdr->MEHDR == CODE_ReqStatsResp)) { 
#ifdef DLANCONFIG_DEBUG
    printf("Calling GetUpdateDevTable\n"); 
#endif
    if(viper_setlocal_neksucceed==0){
      if(!IsSameMAC(pHdr->SA,p_devTable.locBridge.MAC)){
	viper_setlocal_neksucceed=1;
      }
    }
    GetUpdateDevTable(pHdr->SA, ISOLDDEVICE);
    return; 
  }
    
  if( (pHdr->MEHDR == CODE_ReqNetStats)){
#ifdef DLANCONFIG_DEBUG
    printf("Calling GetUpdatePhyDataRates\n"); 
#endif
    GetUpdatePhyDataRates(packet);
    return;
  }
  
  if((pHdr->MEHDR == CODE_SetLocPar) && (pHdr->NETW_CTRL == 0x1)){
#ifdef DLANCONFIG_DEBUG
    printf("Calling GetLocalParExtResp\n"); 
#endif
    GetLocalParExtResp(packet);
    return;
  }

  if( (pHdr->MEHDR == CODE_PrivateVendorSpec) ){    
#ifdef DLANCONFIG_DEBUG
    printf("Calling GetPrivateVendorSpecific\n"); 
#endif
    GetPrivateVendorSpecific(packet);
    return;
  }
  if( (pHdr->MEHDR == CODE_HPSetNEKResp) ){    
#ifdef DLANCONFIG_DEBUG
    printf("Highspeed set local nek succeded\n"); 
#endif
    viper_setlocal_neksucceed=1;
    return;
  }
#ifdef DLANCONFIG_DEBUG
  printf("Packet Ignored\n"); 
#endif
  
}

/*
  return 1 if local bridge have been found
*/
int LocalBridgeAlreadyFound(){
  unsigned char nullMAC[MACSIZE];
  memset(nullMAC,0,MACSIZE);
  return !IsSameMAC(nullMAC,p_devTable.locBridge.MAC);
}

/*
 *   Parse local bridge parameter response
 */
void GetLocalParExtResp(void* packet)
{
  struct mme_locParExtResp* pHdr;
  pHdr = (struct mme_locParExtResp*) packet;  
  if(LocalBridgeAlreadyFound()){
    if(memcmp(p_devTable.locBridge.MAC,pHdr->SA,MACSIZE))
      {
	printf("\n[WARNING]: Found another Local dLAN device.\n");
	printf("             This software cannot manage more than one local device!\n");
	printf("             sorry, exiting!\n");
	exit(1);
      }
    else
      if(olddevice_localnek_setting==1){
	memcpy(p_devTable.locBridge.NEK,pHdr->NEK,DECKSIZE);    
	olddevice_localnek_setting=0;
      }
  }
  else{
    memcpy(p_devTable.locBridge.MAC,pHdr->SA,MACSIZE);
    memcpy(p_devTable.locBridge.NEK,pHdr->NEK,DECKSIZE);
    memcpy(p_devTable.locBridge.DEK,pHdr->DEK,DECKSIZE);
    p_devTable.locBridge.devType=ISOLDDEVICE;
  }
}


/*
 *  Update device table with a new mac
 */
void GetUpdateDevTable(unsigned char*  mac, unsigned char devtype){
  int i=0;

  if(IsSameMAC(mac,p_devTable.locBridge.MAC))
    return;

  for(i=0;i<p_devTable.devCounter;i++)
    if(IsSameMAC(p_devTable.list[i].MAC,mac))
      {
	p_devTable.list[i].isHere=1;
	if(devtype==ISVIPER)
	  p_devTable.list[i].devType=devtype;
	return;
      }

#ifdef DLANCONFIG_DEBUG  
  printf("add device: %X%X%X%X%X%X devtype=%d\n",mac[0],mac[1],mac[2],
  	 mac[3],mac[4],mac[5],devtype);
#endif

  memcpy(p_devTable.list[p_devTable.devCounter].MAC,mac,MACSIZE);
  p_devTable.list[p_devTable.devCounter].devType=devtype;
  p_devTable.list[p_devTable.devCounter].isHere=1;
  p_devTable.devCounter++;
  return;
}

/*
 *  Send packet to set local nek
 */
int SendSetLocalNek( char *iname,unsigned char* NEK_str){  
  unsigned char mac_dest_addr[MACSIZE];  
  struct mme_double* pHdr; 
  unsigned char* pKey;
  int retval=0;

#ifdef DLANCONFIG_DEBUG
  printf("SendSetLocalNEK\n");
#endif
 
  pHdr=(struct mme_double*) malloc(MME_MIN_FRAME_LEN-MME_HDR_LEN);
  //  memset(mac_dest_addr,0xff,MACSIZE);
  memcpy(mac_dest_addr,p_devTable.locBridge.MAC,MACSIZE);
  memset(pHdr,0x00,MME_MIN_FRAME_LEN-MME_HDR_LEN) ; 
  
  pHdr->CTRL  = 0x2 ;   
  pHdr->HDR1  = 0x04;   
  pHdr->LEN1  = 0x09;   
  pHdr->EKS   = 0x01;   

  pKey = MD5HomePlugKey(NEK_str);
  memcpy(pHdr->NEK,pKey,DECKSIZE);
  // printf("%s enc. new nek:",__FUNCTION__);
  // PrintEncDNek(pKey);
  // printf("\n");
    
  if(p_devTable.locBridge.devType!=ISVIPER){
    
#ifdef KEEP_MME_SETCHAR
    /* change characteristics.:  removed to fix bug 1082....*/
    pHdr->HDR2      = 0x1F;
    pHdr->LEN2      = 0x03;
    pHdr->TXCHAR[0] = 0x58;   
    pHdr->TXCHAR[1] = 0x84;
#ifdef KEEP_ENABLE_BACKPRESSURE
    pHdr->TXCHAR[1] |= 0x02;
#endif   
    pHdr->TXCHAR[2] = 0x01;     
    retval=SendPacket(iname,mac_dest_addr,pHdr,MME_MIN_FRAME_LEN-MME_HDR_LEN); 
#else
    retval=SendPacket(iname,mac_dest_addr,pHdr,MME_MIN_FRAME_LEN-MME_HDR_LEN-5);
#endif
  }
  else
    retval=SendPacket(iname,mac_dest_addr,pHdr,MME_MIN_FRAME_LEN-MME_HDR_LEN-5);
  free(pHdr); 
  return retval;	  
}

/*
 * set remote viper nek
*/
int SendSetRemoteViperNEK( char *iname,unsigned char* NEK_str,unsigned char* DEK_str)
{
  unsigned char mac_dest_addr[MACSIZE];  
  MAC_MME_MSG_t* pHdr1;

  // fix bug 1348  mme_single_t* pHdr2;
  mme_setremoteviper_t* pHdr2;
  unsigned char* pnKey;
  unsigned char* prKey;
  int retval=0;
  int dwLen=0;
  int framesize=MME_MIN_FRAME_LEN-MME_HDR_LEN;


#ifdef DLANCONFIG_DEBUG
  printf("SendSetRemoteViperNEK\n");
#endif

  if(sizeof(MAC_MME_MSG_t) > framesize)
    framesize = sizeof(MAC_MME_MSG_t);
  pHdr1=(MAC_MME_MSG_t*) malloc(framesize);
  //  memset(mac_dest_addr,0xff,MACSIZE);
  memset(mac_dest_addr,0xff,MACSIZE);
  memset(pHdr1,0x00,framesize); 

  //  CMACAddress curAddr;
  if(DEK_str)
  {
    pHdr1->mu8Ctrl = 1;
    pHdr1->mu8Header = 0x14;
    pHdr1->mu8Len = 10;
    pHdr1->mu8MsgID = 1;
    pHdr1->mu8EKS = 0;
    prKey = MD5HomePlugKey(DEK_str);
    memcpy(pHdr1->mu8DEK, prKey ,DECKSIZE);
    dwLen = 13;
  }
  // fix bug 1348  pHdr2 = (mme_single_t*)(pHdr1->muData);
  pHdr2 = (mme_setremoteviper_t*)(pHdr1->muData);
  pHdr2->Ethertype =htons(ETHERTYPE_INTELLON);
  // fix bug 1348    pHdr2->CTRL = 1;
  pHdr2->CTRL = 2;
  pHdr2->HDR = 0x04;
  pHdr2->LEN = 9;
  pHdr2->EKS = 1;
  pnKey = MD5HomePlugKey(NEK_str);
  memcpy(pHdr2->NEK,pnKey,DECKSIZE);
  
  // fix bug 1348  added these 3 rows
  pHdr2->HDR2   =   CODE_ReqStats;
  pHdr2->LEN2   =   MME_ReqStatsSize; 
  pHdr2->DUM2   =   0;

  retval=SendPacket(iname,mac_dest_addr,pHdr1,framesize);
  free(pHdr1); 
  return retval; 
}



// Use this function to calculate data rate for user display
inline double Viper_DisplayRate(double dBytesPer336usBlock)
{
  static const double dMaxTrueRate = 6.0;
  static const double dMaxHp1Rate = ((519.0 * 8.0) / (40.0 * 8.4));    // 12.35714286
  static const double dMaxTurboRate = ((2812.0 * 8.0) / (40.0 * 8.4)); // 66.95238095

  double dCalculatedPhyRate = dBytesPer336usBlock * 8.0 / (40.0 * 8.4);

  if(dCalculatedPhyRate <= dMaxTrueRate)
    return dCalculatedPhyRate;
  else if((dCalculatedPhyRate > dMaxTrueRate) && (dCalculatedPhyRate <= dMaxHp1Rate))
    return (((dCalculatedPhyRate - dMaxTrueRate) / (dMaxHp1Rate - dMaxTrueRate)) * (14 - 6)) + 6;
  else
    return (((dCalculatedPhyRate - dMaxHp1Rate) / (dMaxTurboRate - dMaxHp1Rate)) * (85 - 14)) + 14;
}

// Use this function to calculate data rate for user display
inline double HP1_DisplayRate(double dBytes40)
{
  static const double dConstantM = ((588.0 - 38.0) / 481.0);
  static const double dConstantB = (14.0 - dConstantM * (519.0 / 42.0));
  
  return (dConstantM * (dBytes40 / 42.0)) + dConstantB;
}
